@@ -18,7 +18,7 @@ Copyright 2021      IFo Hancroft <contact@ifohancroft.com>
 Copyright 2021      Peter Wu <peterwu@hotmail.com>
 Copyright 2021      Wilson Smith <01wsmith+gh@gmail.com>
 Copyright 2022      Zev Weiss <zev@bewilderbeest.net>
-Copyright 2023      NRK <nrk@disroot.org>
+Copyright 2023-2024 NRK <nrk@disroot.org>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to
@@ -64,21 +64,97 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
     #define PACKAGE_VERSION "unknown"
 #endif
 
+static char defaultOutputFile[] = "%Y-%m-%d-%H%M%S_$wx$h_scrot.$F";
 struct ScrotOptions opt = {
     .quality = 75,
     .compression = 7,
     .lineStyle = LineSolid,
     .lineWidth = 1,
     .lineOpacity = SELECTION_OPACITY_DEFAULT,
-    .lineMode = LINE_MODE_CLASSIC,
     .stackDirection = HORIZONTAL,
-    .monitor = -1,
-    .windowId = None,
-    .outputFile = "%Y-%m-%d-%H%M%S_$wx$h_scrot.png",
+    .outputFile = defaultOutputFile,
     .lineColor = "gray",
 };
 
+enum { /* long opt only */
+    /* ensure these don't collide with single byte opts. */
+    OPT_FORMAT = UCHAR_MAX + 1,
+    OPT_LIST_OPTS,
+};
+static const char stropts[] = "a:bC:cD:d:e:F:fhik::l:M:mn:opq:S:s::t:uvw:Z:z";
+// NOTE: make sure lopts and opt_description indexes are kept in sync
+static const struct option lopts[] = {
+    {"autoselect",      required_argument,  NULL,   'a'},
+    {"border",          no_argument,        NULL,   'b'},
+    {"class",           required_argument,  NULL,   'C'},
+    {"count",           no_argument,        NULL,   'c'},
+    {"display",         required_argument,  NULL,   'D'},
+    {"delay",           required_argument,  NULL,   'd'},
+    {"exec",            required_argument,  NULL,   'e'},
+    {"file",            required_argument,  NULL,   'F'},
+    {"freeze",          no_argument,        NULL,   'f'},
+    {"help",            no_argument,        NULL,   'h'},
+    {"ignorekeyboard",  no_argument,        NULL,   'i'},
+    {"stack",           optional_argument,  NULL,   'k'},
+    {"line",            required_argument,  NULL,   'l'},
+    {"monitor",         required_argument,  NULL,   'M'},
+    {"multidisp",       no_argument,        NULL,   'm'},
+    {"note",            required_argument,  NULL,   'n'},
+    {"overwrite",       no_argument,        NULL,   'o'},
+    {"pointer",         no_argument,        NULL,   'p'},
+    {"quality",         required_argument,  NULL,   'q'},
+    {"script",          required_argument,  NULL,   'S'},
+    {"select",          optional_argument,  NULL,   's'},
+    {"thumb",           required_argument,  NULL,   't'},
+    {"focused",         no_argument,        NULL,   'u'},
+    /* macquarie dictionary has both spellings */
+    {"focussed",        no_argument,        NULL,   'u'},
+    {"version",         no_argument,        NULL,   'v'},
+    {"window",          required_argument,  NULL,   'w'},
+    {"compression",     required_argument,  NULL,   'Z'},
+    {"silent",          no_argument,        NULL,   'z'},
+    {"format",          required_argument,  NULL, OPT_FORMAT},
+    {"list-options",    optional_argument,  NULL, OPT_LIST_OPTS},
+    {0}
+};
+static const char OPT_DEPRECATED[] = "";
+static const struct option_desc {
+    const char *description, *arg_description;
+} opt_description[] = {
+    /* a */  { "autoselect provided region", "x,y,w,h" },
+    /* b */  { "capture the window borders as well", "" },
+    /* C */  { "capture specified window class", "NAME" },
+    /* c */  { "display a countdown for delay", "" },
+    /* D */  { "capture specified display", "DISPLAY" },
+    /* d */  { "add delay before screenshot", "[b]SEC" },
+    /* e */  { "execute command on saved image", "CMD" },
+    /* F */  { "specify output file", "FILE" },
+    /* f */  { "freeze the screen when -s is used", "" },
+    /* h */  { "display help and exit", "" },
+    /* i */  { "ignore keyboard", "" },
+    /* k */  { "capture overlapped window and join them", "v|h" },
+    /* l */  { "specify the style of the selection line", "STYLE" },
+    /* M */  { "capture monitor", "NUM" },
+    /* m */  { "capture all monitors", "" },
+    /* n */  { OPT_DEPRECATED, OPT_DEPRECATED },
+    /* o */  { "overwrite the output file if needed", "" },
+    /* p */  { "capture the mouse pointer as well", "" },
+    /* q */  { "image quality", "NUM" },
+    /* S */  { OPT_DEPRECATED, OPT_DEPRECATED },
+    /* s */  { "interactively select a region to capture", "OPTS" },
+    /* t */  { "also generate a thumbnail", "% | WxH" },
+    /* u */  { "capture the currently focused window", "" },
+    /* u */  { "capture the currently focused window", "" },
+    /* v */  { "output version and exit", "" },
+    /* w */  { "X window ID to capture", "WID" },
+    /* Z */  { "image compression level", "LVL" },
+    /* z */  { "prevent beeping", "" },
+    /* OPT_FORMAT */     { "specify output file format", "FMT" },
+    /* OPT_LIST_OPTS */  { "list all options", "human|tsv" },
+};
+
 static void showUsage(void);
+static void showOptions(bool human);
 static void showVersion(void);
 static long long optionsParseNumBase(const char *, long long, long long,
     const char *[static 1], int);
@@ -143,25 +219,13 @@ static bool optionsParseIsString(const char *const str)
 
 static void optionsParseStack(const char *optarg)
 {
-    // the suboption it's optional
-    if (!optarg) {
+    if (!optarg /* the suboption is optional */ || strcmp(optarg, "h") == 0)
         opt.stackDirection = HORIZONTAL;
-        return;
-    }
-    const char *value = strchr(optarg, '=');
-
-    if (value)
-        ++value;
-    else
-        value = optarg;
-
-    if (*value == 'v')
+    else if (strcmp(optarg, "v") == 0)
         opt.stackDirection = VERTICAL;
-    else if (*value == 'h')
-        opt.stackDirection = HORIZONTAL;
     else {
         errx(EXIT_FAILURE, "option --stack: Unknown value for suboption '%s'",
-             value);
+             optarg);
     }
 }
 
@@ -172,13 +236,7 @@ static void optionsParseSelection(const char *optarg)
         opt.selection.mode = SELECTION_MODE_CAPTURE;
         return;
     }
-
-    const char *value = strchr(optarg, '=');
-
-    if (value)
-        ++value;
-    else
-        value = optarg;
+    const char *value = optarg;
 
     if (!strncmp(value, SELECTION_MODE_S_CAPTURE, SELECTION_MODE_L_CAPTURE)) {
         opt.selection.mode = SELECTION_MODE_CAPTURE;
@@ -280,7 +338,9 @@ static void optionsParseLine(char *optarg)
                     token[Mode]);
             }
 
-            if (!strcmp(value, "classic"))
+            if (!strcmp(value, "auto"))
+                opt.lineMode = LINE_MODE_AUTO;
+            else if (!strcmp(value, "classic"))
                 opt.lineMode = LINE_MODE_CLASSIC;
             else if (!strcmp(value, "edge"))
                 opt.lineMode = LINE_MODE_EDGE;
@@ -315,48 +375,11 @@ static const char *getPathOfStdout(void)
             return paths[i];
     }
     err(EXIT_FAILURE, "access to stdout failed");
+    return 0; /* silence tcc warning */
 }
 
 void optionsParse(int argc, char *argv[])
 {
-    enum { /* long opt only */
-        /* ensure these don't collude with single byte opts. */
-        OPT_FORMAT = UCHAR_MAX + 1,
-    };
-    static char stropts[] = "a:bC:cD:d:e:F:fhik::l:M:mn:opq:S:s::t:uvw:Z:z";
-    static struct option lopts[] = {
-        {"autoselect",      required_argument,  NULL,   'a'},
-        {"border",          no_argument,        NULL,   'b'},
-        {"class",           required_argument,  NULL,   'C'},
-        {"count",           no_argument,        NULL,   'c'},
-        {"display",         required_argument,  NULL,   'D'},
-        {"delay",           required_argument,  NULL,   'd'},
-        {"exec",            required_argument,  NULL,   'e'},
-        {"file",            required_argument,  NULL,   'F'},
-        {"freeze",          no_argument,        NULL,   'f'},
-        {"help",            no_argument,        NULL,   'h'},
-        {"ignorekeyboard",  no_argument,        NULL,   'i'},
-        {"stack",           optional_argument,  NULL,   'k'},
-        {"line",            required_argument,  NULL,   'l'},
-        {"monitor",         required_argument,  NULL,   'M'},
-        {"multidisp",       no_argument,        NULL,   'm'},
-        {"note",            required_argument,  NULL,   'n'},
-        {"overwrite",       no_argument,        NULL,   'o'},
-        {"pointer",         no_argument,        NULL,   'p'},
-        {"quality",         required_argument,  NULL,   'q'},
-        {"script",          required_argument,  NULL,   'S'},
-        {"select",          optional_argument,  NULL,   's'},
-        {"thumb",           required_argument,  NULL,   't'},
-        {"focused",         no_argument,        NULL,   'u'},
-        /* macquarie dictionary has both spellings */
-        {"focussed",        no_argument,        NULL,   'u'},
-        {"version",         no_argument,        NULL,   'v'},
-        {"window",          required_argument,  NULL,   'w'},
-        {"compression",     required_argument,  NULL,   'Z'},
-        {"silent",          no_argument,        NULL,   'z'},
-        {"format",          required_argument,  NULL, OPT_FORMAT},
-        {0}
-    };
     int optch;
     const char *errmsg;
     bool FFlagSet = false;
@@ -366,6 +389,7 @@ void optionsParse(int argc, char *argv[])
     while ((optch = getopt_long(argc, argv, stropts, lopts, NULL)) != -1) {
         switch (optch) {
         case 'a':
+            opt.mode = MODE_AUTOSEL;
             optionsParseAutoselect(optarg);
             break;
         case 'b':
@@ -408,13 +432,14 @@ void optionsParse(int argc, char *argv[])
             opt.ignoreKeyboard = true;
             break;
         case 'k':
-            opt.stack = true;
+            opt.mode = MODE_STACK;
             optionsParseStack(optarg);
             break;
         case 'l':
             optionsParseLine(optarg);
             break;
         case 'M':
+            opt.mode = MODE_MONITOR;
             opt.monitor = optionsParseNum(optarg, 0, INT_MAX, &errmsg);
             if (errmsg) {
                 errx(EXIT_FAILURE, "option --monitor: '%s' is %s", optarg,
@@ -422,7 +447,7 @@ void optionsParse(int argc, char *argv[])
             }
             break;
         case 'm':
-            opt.multidisp = true;
+            opt.mode = MODE_MULTIDISP;
             break;
         case 'n':
             if (optarg[0] == '\0')
@@ -446,18 +471,20 @@ void optionsParse(int argc, char *argv[])
             opt.script = optarg;
             break;
         case 's':
+            opt.mode = MODE_SELECT;
             optionsParseSelection(optarg);
             break;
         case 't':
             optionsParseThumbnail(optarg);
             break;
         case 'u':
-            opt.focused = true;
+            opt.mode = MODE_FOCUSED;
             break;
         case 'v':
             showVersion();
             break;
         case 'w':
+            opt.mode = MODE_WINDOW;
             opt.windowId = optionsParseNumBase(optarg, None/*0L*/, LONG_MAX, &errmsg, 0);
             if (errmsg) {
                 errx(EXIT_FAILURE, "option --window: '%s' is %s", optarg,
@@ -477,6 +504,16 @@ void optionsParse(int argc, char *argv[])
         case OPT_FORMAT:
             opt.format = optarg;
             break;
+        case OPT_LIST_OPTS:
+            if (optarg == NULL || strcmp(optarg, "human") == 0)
+                showOptions(true);
+            else if (strcmp(optarg, "tsv") == 0)
+                showOptions(false);
+            else {
+                errx(EXIT_FAILURE,
+                    "unknown argument for --list-options: `%s`", optarg);
+            }
+            break;
         default:
             exit(EXIT_FAILURE);
         }
@@ -491,11 +528,27 @@ void optionsParse(int argc, char *argv[])
     for (; *argv; ++argv)
         warnx("ignoring extraneous non-option argument: %s", *argv);
 
+    if (!opt.format) {
+        char *ext;
+        size_t extLength = scrotHaveFileExtension(opt.outputFile, &ext);
+        if (extLength == 0 || opt.outputFile == defaultOutputFile)
+            opt.format = "png";
+        else
+            opt.format = ext+1;
+    }
+
     if (strcmp(opt.outputFile, "-") == 0) {
         opt.overwrite = true;
         opt.thumb = THUMB_DISABLED;
         opt.outputFile = getPathOfStdout();
     }
+
+    size_t outputFileLen = strlen(opt.outputFile);
+    if (outputFileLen == 0)
+        errx(EXIT_FAILURE, "output file cannot be empty");
+    if (opt.outputFile[outputFileLen - 1] == '/')
+        errx(EXIT_FAILURE, "output file cannot be a directory");
+
     if (opt.thumb != THUMB_DISABLED)
         opt.thumbFile = optionsNameThumbnail(opt.outputFile);
 
@@ -509,14 +562,49 @@ void optionsParse(int argc, char *argv[])
 
 static void showUsage(void)
 {
-    fputs("usage:  " /* Check that everything lines up after any changes. */
-        PACKAGE_NAME " [-bcfhimopuvz] [-a X,Y,W,H] [-C NAME] [-D DISPLAY]\n"
-        "              [-d SEC] [-e CMD] [-k OPT] [-l STYLE] [-M NUM] [-q NUM]\n"
-        "              [-s OPTS] [-t % | WxH] [[-F] FILE]\n",
-        stdout);
+    fputs("Usage:  " PACKAGE_NAME " [OPTIONS...] [FILE]\n\n"
+          "A list of options with brief description is given below.\n"
+          "For more detailed description, "
+          "consult the "PACKAGE_NAME"(1) manpage.\n\n", stdout);
+    showOptions(true);
     exit(0);
 }
 
+static void showOptions(bool human)
+{
+    for (size_t i = 0; i < ARRAY_COUNT(opt_description); ++i) {
+        const struct option *o = &lopts[i];
+        const struct option_desc *d = &opt_description[i];
+        if (d->description == OPT_DEPRECATED)
+            continue;
+
+        if (human) {
+            int n = 0;
+            if (o->val <= UCHAR_MAX)
+                n += printf("-%c, ", o->val);
+            n += printf("--%s", o->name);
+            if (o->has_arg == required_argument)
+                n += printf(" <%s>", d->arg_description);
+            else if (o->has_arg == optional_argument)
+                n += printf("[=%s]", d->arg_description);
+            for (; n >= 0 && n < 32; ++n)
+                putchar(' ');
+            printf("%s\n", d->description);
+        } else {
+            printf("%c\t", o->val <= UCHAR_MAX ? o->val : ' ');
+            printf("%s\t", o->name);
+            if (o->has_arg == required_argument)
+                printf("R:%s\t", d->arg_description);
+            else if (o->has_arg == optional_argument)
+                printf("O:%s\t", d->arg_description);
+            else
+                printf("N:\t");
+            printf("%s\n", d->description);
+        }
+    }
+    exit(EXIT_SUCCESS);
+}
+
 static void showVersion(void)
 {
     puts(PACKAGE_NAME " version " PACKAGE_VERSION);
@@ -528,13 +616,13 @@ static char *optionsNameThumbnail(const char *name)
     const ptrdiff_t nameLength = strlen(name);
     const char thumbSuffix[] = "-thumb";
     Stream ret = {0};
-    const char *const extension = strrchr(name, '.');
-    const ptrdiff_t baseNameLength = extension ? extension-name : nameLength;
+    char *extension;
+    size_t extLength = scrotHaveFileExtension(name, &extension);
+    const ptrdiff_t baseNameLength = nameLength - extLength;
 
     streamMem(&ret, name, baseNameLength);
     streamMem(&ret, thumbSuffix, sizeof(thumbSuffix)-1);
-    if (extension)
-        streamMem(&ret, extension, nameLength - baseNameLength);
+    streamMem(&ret, extension, extLength);
     streamChar(&ret, '\0');
 
     return ret.buf;
@@ -562,8 +650,6 @@ void optionsParseAutoselect(char *optarg)
                 errmsg);
         }
         i++;
-
-        opt.autoselect = true;
     }
     if (i < 4)
         errx(EXIT_FAILURE, "option --autoselect: too few dimensions");
