@@ -65,12 +65,11 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include "note.h"
 #include "options.h"
 #include "scrot.h"
-#include "slist.h"
 #include "util.h"
 
 static void initXAndImlib(const char *, int);
 static void uninitXAndImlib(void);
-static size_t scrotHaveFileExtension(const char *, char **);
+static void scrotSaveImage(const char *);
 static Imlib_Image scrotGrabFocused(void);
 static void applyFilterIfRequired(void);
 static Imlib_Image scrotGrabAutoselect(void);
@@ -86,9 +85,10 @@ static char *imPrintf(const char *, struct tm *, const char *, const char *,
 static char *scrotGetWindowName(Window);
 static Window scrotGetClientWindow(Display *, Window);
 static Window scrotFindWindowByProperty(Display *, const Window, const Atom);
-static Imlib_Image stalkImageConcat(ScrotList *, const enum Direction);
+static Imlib_Image stalkImageConcat(Imlib_Image *, size_t, const enum Direction);
 static int findWindowManagerFrame(Window *const, int *const);
 static Imlib_Image scrotGrabWindowById(Window const window);
+static void scrotGrabMousePointer(Imlib_Image, const int, const int);
 
 /* X11 stuff */
 Display *disp;
@@ -98,12 +98,10 @@ Screen *scr;
 
 int main(int argc, char *argv[])
 {
-    Imlib_Image image;
+    Imlib_Image image = NULL;
     Imlib_Image thumbnail;
-    Imlib_Load_Error imErr;
     char *filenameIM = NULL;
     char *filenameThumb = NULL;
-    char *haveExtension = NULL;
     struct timespec timeStamp;
     struct tm *tm;
 
@@ -116,26 +114,28 @@ int main(int argc, char *argv[])
 
     initXAndImlib(opt.display, 0);
 
-    if (opt.selection.mode & SELECTION_MODE_ANY)
+    if (opt.mode == MODE_SELECT)
         image = scrotSelectionSelectMode();
     else {
 
         scrotDoDelay();
 
-        if (opt.focused)
+        if (opt.mode == MODE_FOCUSED)
             image = scrotGrabFocused();
-        else if (opt.multidisp)
+        else if (opt.mode == MODE_MULTIDISP)
             image = scrotGrabShotMulti();
-        else if (opt.stack)
+        else if (opt.mode == MODE_STACK)
             image = scrotGrabStackWindows();
-        else if (opt.monitor != -1)
+        else if (opt.mode == MODE_MONITOR)
             image = scrotGrabShotMonitor();
-        else if (opt.autoselect)
+        else if (opt.mode == MODE_AUTOSEL)
             image = scrotGrabAutoselect();
-        else if (opt.windowId != None)
+        else if (opt.mode == MODE_WINDOW)
             image = scrotGrabWindowById(opt.windowId);
-        else
+        else if (opt.mode == MODE_SCREEN)
             image = scrotGrabShot();
+        else
+            scrotAssert(!"unreachable");
     }
     if (!image)
         errx(EXIT_FAILURE, "no image grabbed");
@@ -155,13 +155,7 @@ int main(int argc, char *argv[])
         scrotNoteDraw(image);
 
     imlib_context_set_image(image);
-    if (opt.format) {
-        imlib_image_set_format(opt.format);
-    } else {
-        scrotHaveFileExtension(opt.outputFile, &haveExtension);
-        if (!haveExtension)
-            imlib_image_set_format("png");
-    }
+    imlib_image_set_format(opt.format);
     imlib_image_attach_data_value("quality", NULL, opt.quality, NULL);
     imlib_image_attach_data_value("compression", NULL, opt.compression, NULL);
 
@@ -170,9 +164,7 @@ int main(int argc, char *argv[])
 
     applyFilterIfRequired();
 
-    imlib_save_image_with_error_return(filenameIM, &imErr);
-    if (imErr)
-        errx(EXIT_FAILURE, "Saving to file %s failed", filenameIM);
+    scrotSaveImage(filenameIM);
 
     if (opt.thumb != THUMB_DISABLED) {
         int cwidth, cheight;
@@ -207,18 +199,12 @@ int main(int argc, char *argv[])
             errx(EXIT_FAILURE, "unable to create thumbnail");
         } else {
             imlib_context_set_image(thumbnail);
-            if (opt.format)
-                imlib_image_set_format(opt.format);
-            else if (!haveExtension)
-                imlib_image_set_format("png");
+            imlib_image_set_format(opt.format);
 
             filenameThumb = imPrintf(opt.thumbFile, tm, NULL, NULL, thumbnail);
             scrotCheckIfOverwriteFile(&filenameThumb);
-            imlib_save_image_with_error_return(filenameThumb, &imErr);
+            scrotSaveImage(filenameThumb);
             imlib_free_image_and_decache();
-
-            if (imErr)
-                errx(EXIT_FAILURE, "Saving thumbnail %s failed", filenameThumb);
         }
     }
     if (opt.exec)
@@ -263,6 +249,7 @@ static void initXAndImlib(const char *dispStr, int screenNumber)
     imlib_context_set_colormap(cm);
     imlib_context_set_color_modifier(NULL);
     imlib_context_set_operation(IMLIB_OP_COPY);
+    imlib_set_cache_size(0);
 }
 
 /* atexit register func. */
@@ -279,16 +266,45 @@ static void uninitXAndImlib(void)
     }
 }
 
-static size_t scrotHaveFileExtension(const char *filename, char **ext)
+static void scrotSaveImage(const char *filename)
 {
-    char *s = strrchr(filename, '.');
-
-    if (s && s[1] != '\0') {
-        *ext = s;
-        return strlen(s);
+    Imlib_Load_Error imErr;
+    imlib_save_image_with_error_return(filename, &imErr);
+    if (imErr) {
+        const char *colon = "", *errmsg = "", imlibPrefix[] = "Imlib2: ";
+#if defined(IMLIB2_VERSION)
+#if IMLIB2_VERSION >= IMLIB2_VERSION_(1, 10, 0)
+        colon = ": ";
+        errmsg = imlib_strerror(imlib_get_error());
+#endif
+#endif
+        if (strncmp(errmsg, imlibPrefix, sizeof(imlibPrefix) - 1) == 0)
+            errmsg += sizeof(imlibPrefix) - 1;
+        errx(EXIT_FAILURE, "failed to save image: %s%s%s",
+            filename, colon, errmsg);
     }
+}
 
-    return 0;
+size_t scrotHaveFileExtension(const char *filename, char **ext)
+{
+    const char *basename = strrchr(filename, '/');
+    if (!basename)
+        basename = filename;
+    basename += *basename == '/';
+    basename += *basename == '.'; /* dot files are not extensions */
+    char *s = strrchr(basename, '.');
+    *ext = (s && s[1] != '\0') ? s : "";
+    return strlen(*ext);
+}
+
+Imlib_Image scrotGrabRectAndPointer(int x, int y, int w, int h)
+{
+    Imlib_Image im = imlib_create_image_from_drawable(0, x, y, w, h, 1);
+    if (!im)
+        errx(EXIT_FAILURE, "failed to grab image");
+    if (opt.pointer)
+        scrotGrabMousePointer(im, x, y);
+    return im;
 }
 
 static Imlib_Image scrotGrabWindowById(Window const window)
@@ -299,9 +315,7 @@ static Imlib_Image scrotGrabWindowById(Window const window)
     if (!scrotGetGeometry(window, &rx, &ry, &rw, &rh))
         return NULL;
     scrotNiceClip(&rx, &ry, &rw, &rh);
-    im = imlib_create_image_from_drawable(0, rx, ry, rw, rh, 1);
-    if (opt.pointer)
-        scrotGrabMousePointer(im, rx, ry);
+    im = scrotGrabRectAndPointer(rx, ry, rw, rh);
     clientWindow = window;
     return im;
 }
@@ -316,15 +330,10 @@ static Imlib_Image scrotGrabFocused(void)
 
 static Imlib_Image scrotGrabAutoselect(void)
 {
-    Imlib_Image im = NULL;
     int rx = opt.autoselectX, ry = opt.autoselectY, rw = opt.autoselectW,
         rh = opt.autoselectH;
-
     scrotNiceClip(&rx, &ry, &rw, &rh);
-    im = imlib_create_image_from_drawable(0, rx, ry, rw, rh, 1);
-    if (opt.pointer)
-        scrotGrabMousePointer(im, rx, ry);
-    return im;
+    return scrotGrabRectAndPointer(rx, ry, rw, rh);
 }
 
 void scrotDoDelay(void)
@@ -332,12 +341,17 @@ void scrotDoDelay(void)
     if (!opt.delay)
         return;
     if (opt.countdown) {
-        dprintf(STDERR_FILENO, "Taking shot in ");
+        fputs("Taking shot in ", stderr);
         for (int i = opt.delay; i > 0; i--) {
-            dprintf(STDERR_FILENO, "%d.. ", i);
+            /* Illumos doesn't have dprintf():
+             * https://www.illumos.org/issues/1609
+             */
+            fprintf(stderr, "%d.. ", i);
+            fflush(stderr);
             opt.delayStart = scrotSleepFor(opt.delayStart, 1000);
         }
-        dprintf(STDERR_FILENO, "0.\n");
+        fputs("0.\n", stderr);
+        fflush(stderr);
     } else {
         scrotSleepFor(opt.delayStart, opt.delay * 1000);
     }
@@ -505,7 +519,7 @@ Window scrotGetWindow(Display *display, Window window, int x, int y)
     return target;
 }
 
-void scrotGrabMousePointer(Imlib_Image image, const int xOffset,
+static void scrotGrabMousePointer(Imlib_Image image, const int xOffset,
     const int yOffset)
 {
     XFixesCursorImage *xcim = XFixesGetCursorImage(disp);
@@ -559,44 +573,34 @@ static void scrotCheckIfOverwriteFile(char **filename)
         return;
 
     const size_t maxCounter = 999;
-    size_t counter = 0;
-    char *ext = NULL;
-    size_t extLength = 0;
+    char fmt[5]; // _000 + NUL byte
     const size_t slen = strlen(*filename);
-    size_t nalloc = slen + 4 + 1; // _000 + NUL byte
-    char fmt[5];
-    char *newName = NULL;
+    const size_t nalloc = slen + sizeof(fmt);
 
-    extLength = scrotHaveFileExtension(*filename, &ext);
+    char *ext;
+    size_t extLength = scrotHaveFileExtension(*filename, &ext);
 
-    if (ext)
-        nalloc += extLength; // .ext
-
-    newName = ecalloc(nalloc, sizeof(*newName));
-    memcpy(newName, *filename, slen);
+    char *newName = ecalloc(nalloc, sizeof(*newName));
+    memcpy(newName, *filename, slen - extLength);
+    char *ptr = newName + (slen - extLength);
 
+    size_t counter = 0;
     do {
-        char *ptr = newName + slen;
-
         snprintf(fmt, sizeof(fmt), "_%03zu", counter++);
-
-        if (ext) {
-            ptr -= extLength;
-            memcpy(ptr, fmt, sizeof(fmt));
-            memcpy(ptr + sizeof(fmt) - 1, ext, extLength);
-        } else
-            memcpy(ptr, fmt, sizeof(fmt));
+        memcpy(ptr, fmt, sizeof(fmt));
+        memcpy(ptr + sizeof(fmt) - 1, ext, extLength);
     } while ((counter < maxCounter) && !access(newName, F_OK));
 
     scrotAssert(newName[nalloc - 1] == '\0');
 
-    free(*filename);
-    *filename = newName;
-
     if (counter == maxCounter) {
         errx(EXIT_FAILURE, "scrot can no longer generate new file names.\n"
             "The last attempt is %s", newName);
     }
+
+    warnx("`%s` already exists, attempting `%s` instead", *filename, newName);
+    free(*filename);
+    *filename = newName;
 }
 
 static int scrotMatchWindowClassName(Window target)
@@ -619,17 +623,9 @@ static int scrotMatchWindowClassName(Window target)
 
 static Imlib_Image scrotGrabShot(void)
 {
-    Imlib_Image im;
-
     if (!opt.silent)
         XBell(disp, 0);
-
-    im = imlib_create_image_from_drawable(0, 0, 0, scr->width,
-        scr->height, 1);
-    if (opt.pointer)
-        scrotGrabMousePointer(im, 0, 0);
-
-    return im;
+    return scrotGrabRectAndPointer(0, 0, scr->width, scr->height);
 }
 
 static void scrotExecApp(Imlib_Image image, struct tm *tm, char *filenameIM,
@@ -637,11 +633,10 @@ static void scrotExecApp(Imlib_Image image, struct tm *tm, char *filenameIM,
 {
     char *execStr = imPrintf(opt.exec, tm, filenameIM, filenameThumb, image);
     int ret = system(execStr);
-
     if (ret == -1)
-        warn("The child process could not be created");
-    else if (WIFEXITED(ret) && WEXITSTATUS(ret) == 127)
-        warnx("scrot could not execute the command: %s", execStr);
+        err(EXIT_FAILURE, "The child process could not be created");
+    else if (!WIFEXITED(ret) || WEXITSTATUS(ret) != 0)
+        exit(69 /* EX_UNAVAILABLE */ );
     free(execStr);
 }
 
@@ -651,16 +646,29 @@ static char *imPrintf(const char *str, struct tm *tm, const char *filenameIM,
     char buf[20];
     Stream ret = {0};
     long hostNameMax = 0;
-    char strf[4096];
     char *tmp;
     struct stat st;
 
-    const size_t strfLen = strftime(strf, sizeof(strf), str, tm);
+    /* prepend a useless '.' to distinguish between valid 0 return vs 0 return
+     * due to small buffer size.
+     */
+    Stream fmt = {0};
+    streamChar(&fmt, '.');
+    streamStr(&fmt, str);
+    streamChar(&fmt, '\0');
+
+    size_t strfLen, strfSize = 32;
+    char *strf = NULL;
+    do {
+        strfSize *= 2;
+        strf = erealloc(strf, strfSize);
+    } while ((strfLen = strftime(strf, strfSize, fmt.buf, tm)) == 0);
+    free(fmt.buf);
     if (strfLen == 0)
         errx(EXIT_FAILURE, "strftime returned 0");
 
     imlib_context_set_image(im);
-    for (const char *c = strf, *end = strf + strfLen; c < end; ++c) {
+    for (const char *c = strf + 1, *end = strf + strfLen; c < end; ++c) {
         if (*c == '$' && (c + 1) < end) {
             c++;
             switch (*c) {
@@ -676,6 +684,9 @@ static char *imPrintf(const char *str, struct tm *tm, const char *filenameIM,
                 ret.off += strlen(target);
                 scrotAssert(ret.buf[ret.off] == '\0');
                 break;
+            case 'F':
+                streamStr(&ret, opt.format);
+                break;
             case 'f':
                 if (filenameIM)
                     streamStr(&ret, filenameIM);
@@ -704,10 +715,7 @@ static char *imPrintf(const char *str, struct tm *tm, const char *filenameIM,
             case 's':
                 if (filenameIM) {
                     if (!stat(filenameIM, &st)) {
-                        int size;
-
-                        size = st.st_size;
-                        snprintf(buf, sizeof(buf), "%d", size);
+                        snprintf(buf, sizeof(buf), "%jd", (intmax_t)st.st_size);
                         streamStr(&ret, buf);
                     } else
                         streamStr(&ret, "[err]");
@@ -751,6 +759,7 @@ static char *imPrintf(const char *str, struct tm *tm, const char *filenameIM,
             streamChar(&ret, *c);
         }
     }
+    free(strf);
     streamChar(&ret, '\0');
     return ret.buf;
 }
@@ -838,8 +847,8 @@ static Imlib_Image scrotGrabStackWindows(void)
 {
     if (XGetSelectionOwner(disp, XInternAtom(disp, "_NET_WM_CM_S0", False))
         == None) {
-        errx(EXIT_FAILURE, "option --stack: Composite Manager is not running,"
-            " required to use this option.");
+        warnx("option --stack: Composite Manager is not running,"
+            " some windows may overlap.");
     }
 
     unsigned long numberItemsReturn;
@@ -868,7 +877,11 @@ static Imlib_Image scrotGrabStackWindows(void)
             EWMH_CLIENT_LIST);
     }
 
-    initializeScrotList(images);
+    size_t imagesCount = 0;
+    Imlib_Image *images = erealloc(NULL, numberItemsReturn * sizeof(*images));
+
+    if (!XCompositeQueryVersion(disp, &(int){0}, &(int){0}))
+        errx(EXIT_FAILURE, "XCompositeQueryVersion() failed");
 
     XCompositeRedirectSubwindows(disp, root, CompositeRedirectAutomatic);
 
@@ -893,15 +906,20 @@ static Imlib_Image scrotGrabStackWindows(void)
                 "option --stack: Failed XGetImage: Window id 0x%lx", win);
         }
 
-        im = imlib_create_image_from_ximage(ximage, NULL, attr.x, attr.y,
+        im = imlib_create_image_from_ximage(ximage, NULL, 0, 0,
             attr.width, attr.height, 1);
+        if (!im) {
+            errx(EXIT_FAILURE, "option --stack: "
+                "Failed to create Imlib2 image: Window id 0x%lx", win);
+        }
 
-        XFree(ximage);
+        XDestroyImage(ximage);
 
-        appendToScrotList(images, im);
+        images[imagesCount++] = im;
     }
+    XFree(propReturn);
 
-    return stalkImageConcat(&images, opt.stackDirection);
+    return stalkImageConcat(images, imagesCount, opt.stackDirection);
 }
 
 static Imlib_Image scrotGrabShotMulti(void)
@@ -915,8 +933,8 @@ static Imlib_Image scrotGrabShotMulti(void)
     char *subDisp;
     char newDisp[255];
     Imlib_Image ret = NULL;
-
-    initializeScrotList(images);
+    size_t imagesCount = 0;
+    Imlib_Image *images = erealloc(NULL, screens * sizeof(*images));
 
     subDisp = estrdup(DisplayString(disp));
 
@@ -931,12 +949,14 @@ static Imlib_Image scrotGrabShotMulti(void)
         initXAndImlib(newDisp, i);
         ret = imlib_create_image_from_drawable(0, 0, 0, scr->width,
             scr->height, 1);
+        if (!ret)
+            errx(EXIT_FAILURE, "failed to grab image");
 
-        appendToScrotList(images, ret);
+        images[imagesCount++] = ret;
     }
     free(subDisp);
 
-    return stalkImageConcat(&images, HORIZONTAL);
+    return stalkImageConcat(images, imagesCount, HORIZONTAL);
 }
 
 static Imlib_Image scrotGrabShotMonitor(void)
@@ -960,32 +980,27 @@ static Imlib_Image scrotGrabShotMonitor(void)
 
     scrotAssert(screens); /* silence clang-tidy */
     XineramaScreenInfo *mon = &screens[opt.monitor];
-
-    /* Hack: pretend we were invoked in autoselect mode */
-    opt.autoselectX = mon->x_org;
-    opt.autoselectY = mon->y_org;
-    opt.autoselectW = mon->width;
-    opt.autoselectH = mon->height;
-
+    int x = mon->x_org, y = mon->y_org, w = mon->width, h = mon->height;
     XFree(screens);
 
-    return scrotGrabAutoselect();
+    scrotNiceClip(&x, &y, &w, &h);
+    return scrotGrabRectAndPointer(x, y, w, h);
 }
 
-static Imlib_Image stalkImageConcat(ScrotList *images, const enum Direction dir)
+static Imlib_Image stalkImageConcat(
+    Imlib_Image *images, size_t imagesCount, const enum Direction dir)
 {
-    if (isEmptyScrotList(images))
+    if (imagesCount == 0)
         return NULL;
 
     int total = 0, max = 0;
     int x = 0, y = 0, w , h;
     Imlib_Image ret, im;
-    ScrotListNode *image = NULL;
 
     const bool vertical = (dir == VERTICAL) ? true : false;
 
-    forEachScrotList(images, image) {
-        im = (Imlib_Image) image->data;
+    for (size_t i = 0; i < imagesCount; ++i) {
+        im = images[i];
         imlib_context_set_image(im);
         h = imlib_image_get_height();
         w = imlib_image_get_width();
@@ -1012,22 +1027,21 @@ static Imlib_Image stalkImageConcat(ScrotList *images, const enum Direction dir)
     imlib_context_set_color(0, 0, 0, 255);
     imlib_image_fill_rectangle(0, 0, w, h);
 
-    image = firstScrotList(images);
-    while (image) {
-        im = (Imlib_Image) image->data;
+    imlib_context_set_anti_alias(0);
+    imlib_context_set_dither(1);
+    imlib_context_set_blend(0);
+
+    for (size_t i = 0; i < imagesCount; ++i) {
+        im = images[i];
         imlib_context_set_image(im);
         h = imlib_image_get_height();
         w = imlib_image_get_width();
         imlib_context_set_image(ret);
-        imlib_context_set_anti_alias(0);
-        imlib_context_set_dither(1);
-        imlib_context_set_blend(0);
-        imlib_context_set_angle(0);
         imlib_blend_image_onto_image(im, 0, 0, 0, w, h, x, y, w, h);
         (!vertical) ? (x += w) : (y += h);
         imlib_context_set_image(im);
         imlib_free_image_and_decache();
-        nextAndFreeScrotList(image);
     }
+    free(images);
     return ret;
 }
