@@ -51,6 +51,7 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #define SELECTION_MODE_L_BLUR 4
 
 enum LineMode {
+    LINE_MODE_AUTO,
     LINE_MODE_CLASSIC,
     LINE_MODE_EDGE,
 };
@@ -83,15 +84,20 @@ typedef struct SelectionMode {
     const char *fileName;
 } SelectionMode;
 
-struct SelectionClassic;
-struct SelectionEdge;
+struct SelectionClassic {
+    XGCValues gcValues;
+    GC gc;
+};
+struct SelectionEdge {
+    Window wndDraw;
+};
 
 struct Selection {
     Cursor curCross, curAngleNW, curAngleNE, curAngleSW, curAngleSE;
 
     struct SelectionRect rect;
-    struct SelectionClassic *classic;
-    struct SelectionEdge *edge;
+    struct SelectionClassic classic;
+    struct SelectionEdge edge;
 
     void (*create)(void);
     void (*destroy)(void);
@@ -99,17 +105,10 @@ struct Selection {
     void (*motionDraw)(int, int, int, int);
 };
 
-struct Selection **selectionGet(void);
+extern struct Selection selection;
+
 void selectionCalculateRect(int, int, int, int);
-void scrotSelectionCreate(void);
-void scrotSelectionDestroy(void);
-void scrotSelectionDraw(void);
-void scrotSelectionMotionDraw(int, int, int, int);
-struct SelectionRect *scrotSelectionGetRect(void);
 void scrotSelectionGetLineColor(XColor *);
-Status scrotSelectionCreateNamedColor(const char *, XColor *);
-void scrotSelectionSetDefaultColorLine(void);
-bool scrotSelectionGetUserSel(struct SelectionRect *);
 Imlib_Image scrotSelectionSelectMode(void);
 
 #endif /* !defined(H_SCROT_SELECTION) */
